// treeasis.c source : Main source file for the parser implementation
//

#ifdef _WIN32
#include "stdafx.h"
#endif

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <BipCpi.h>
#include <BipCni.h>
#include <BipCci.h>
#include "ParserUtils.h"

#ifdef _WIN32
#include "NT_msg/treeasis_msg.h"
#else
#include "AIX_msg/treeasis_msg.h"
#endif

#ifdef _WIN32
#define LOGFILE "C:\\TREEASIS.TXT"
#else
#define LOGFILE "/tmp/treeasis.txt"
#endif

// This may be fixed in a later release of MQSI V2 ...
typedef struct CciByteArray CciByteArray;
static	FILE	*logFile = 0;
static	char	sccsid[] = "@(#) MQSI V2.0.1 TREEASIS V1.1, kolban@us.ibm.com";

#ifdef _WIN32
void __declspec(dllexport) *bipGetParserFactory();
HANDLE	g_hModule;
BOOL APIENTRY DllMain( HANDLE hModule, 
                       DWORD  ul_reason_for_call, 
                       LPVOID lpReserved
					 )
{
	g_hModule = hModule;
    return TRUE;
}
#endif


///////////////////////////////////////////////////////////
// ErrorToString
char *ErrorToString(int iRc)
{
	static char	message[256];
	switch(iRc)
	{
	case CCI_PARSER_NAME_TOO_LONG:
		return("Parse name too long");
	default:
		sprintf(message, "Error: Unknown code: %d", iRc);
		return(message);
	}
}

///////////////////////////////////////////////////////////
// Log
void Log(char *szFormat, ...)
{
#ifndef _DEBUG
	return;
#else

	va_list	list;
	va_start(list, szFormat);
	if (logFile == NULL)
	{
		logFile = fopen(LOGFILE, "a");
		if (logFile == NULL)
		{
			return;
		}
	}
	vfprintf(logFile, szFormat, list);
	fflush(logFile);
	va_end(list);
#endif
}



///////////////////////////////////////////////////////////
// CreateContext
void *CreateContext(CciParser *parser)
{
	struct TreeAsIsContext *pTreeAsIsContext;

	Log(">> CreateContext\n");
	pTreeAsIsContext = (struct TreeAsIsContext *)malloc(sizeof(struct TreeAsIsContext));
	pTreeAsIsContext->iSize = 256;
	pTreeAsIsContext->pData = (CciChar *)malloc(sizeof(CciChar) * pTreeAsIsContext->iSize + 1);
	Log("<< CreateContext\n");
	return(pTreeAsIsContext);
}

///////////////////////////////////////////////////////////
// Delete Context
void DeleteContext(CciParser *parser, CciContext *context)
{
	struct TreeAsIsContext *pTreeAsIsContext;


	Log(">> Entered DeleteContext\n");
	pTreeAsIsContext = (struct TreeAsIsContext *)context;
	if (pTreeAsIsContext != NULL)
	{
		if (pTreeAsIsContext->pData != NULL)
		{
			free(pTreeAsIsContext->pData);
		}
		free(pTreeAsIsContext);
	}
	Log("<< Entered DeleteContext\n");
}

///////////////////////////////////////////////////////////
// ParseBuffer
int ParseBuffer(CciParser	*parser,
			    CciContext	*context,
			    int         iOffset)
{
	int		iRc;
	cciThrowException(
		&iRc,						// Return code
		CCI_PARSER_EXCEPTION,		// Type of exception
		__FILE__,					// File name
		__LINE__,					// Exception source line number
		"ParseBuffer",				// Function that threw exception
		L"MQSIV2_TREEASIS",			// Message source
		MSG_INVALID_PARSER_USAGE,	// Message number
		"Attempt to use TREEASIS to parse incoming data",
		(char *)0					// Parameters
	);
	return(0);
}

///////////////////////////////////////////////////////////
// ParseFirstChild
void ParseFirstChild(CciParser	*parser,
					 CciContext	*context,
					 CciElement	*currentElement)
{
	int		iRc;
	CciElement	*rootElement;
	Log(">> ParseFirstChild() element=%p\n", currentElement);
	rootElement = cpiRootElement(&iRc, parser);
	cpiSetElementCompleteNext(&iRc, rootElement, 1);
	cpiSetElementCompletePrevious(&iRc, rootElement, 1);
	Log("<< ParseFirstChild()\n");
}

void ParseLastChild(CciParser	*parser,
					   CciContext	*context,
					   CciElement	*currentElement)
{
	Log(">> ParseLastChild()\n");
}

void ParsePreviousSibling(CciParser		*parser,
						  CciContext	*context,
						  CciElement	*currentElement)
{
	Log(">> ParsePreviousSibling()\n");
}


void ParseNextSibling(CciParser	*parser,
						CciContext	*context,
						CciElement	*currentElement)
{
	Log(">> ParseNextSibling()\n");
}

///////////////////////////////////////////////////////////
// ProcessNode
void ProcessNode(struct TreeAsIsContext *pTreeAsIsContext,
				 CciParser	*parser,
				 CciElement	*node,
				 CciByteArray *tempByteArray)
{
	int				iRc;
	char			*pValue;
	CciInt			intVal;
	CciElement		*child;
	CciElementType	elementType;
	CciValueType	valueType;
	CciSize			byteArraySize;

	elementType = cpiElementType(&iRc, node);
	if ((elementType & CCI_ELEMENT_TYPE_MASK_GENERIC) == 
		CCI_ELEMENT_TYPE_NAME_VALUE ||
		(elementType & CCI_ELEMENT_TYPE_MASK_GENERIC) == 
		CCI_ELEMENT_TYPE_VALUE)
	{
		valueType = cniElementValueType(&iRc, node);
		switch(valueType)
		{
		case CCI_VALUE_TYPE_CHAR:
			pValue = GetStringValue(pTreeAsIsContext, parser, node);
			cpiAppendToBuffer(&iRc, parser, pValue, strlen(pValue));
			free(pValue);
			break;

		case CCI_VALUE_TYPE_INTEGER:
			intVal = cpiElementIntegerValue(&iRc, node);
			cpiAppendToBuffer(&iRc, parser, (char *)&intVal, sizeof(intVal));
			break;

		case CCI_VALUE_TYPE_BYTEARRAY:
			byteArraySize = cpiElementByteArrayValue(
				&iRc,
				node,
				tempByteArray);
			if (iRc == CCI_BUFFER_TOO_SMALL)
			{
				tempByteArray->size = byteArraySize;
				free(tempByteArray->pointer);
				tempByteArray->pointer = malloc(tempByteArray->size);
				byteArraySize = cpiElementByteArrayValue(
					&iRc,
					node,
					tempByteArray);
			}
			cpiAppendToBuffer(&iRc,
				parser,
				(char *)tempByteArray->pointer,
				byteArraySize);
			break;

		default:
			Log("Unknown value type: %d\n", valueType);
			break;
		}
	}
	child = cpiFirstChild(&iRc, node);
	while(child != CCI_NULL_ADDR)
	{
		ProcessNode(pTreeAsIsContext, parser, child, tempByteArray);
		child = cpiNextSibling(&iRc, child);
	}
}

///////////////////////////////////////////////////////////
// WriteBuffer
//
// In the WriteBuffer algorithm, it is possible that we will encounter
// byte array data.  This implementation of the parser allocates an
// initial 100 bytes for such a data structure.  If the byte array can
// be retrieved into this buffer, great.  If not, the buffer will be
// deleted and recreated large enough for the byte array located.
// The buffer of this new, larger size will be maintained in case
// it can be reused for a subsequent byte array later on in the
// reconstitution of the binary data.
//
int WriteBuffer(CciParser	*parser,
				CciContext	*context)
{
	int				iRc;
	CciElement		*rootElement;
	CciByteArray	tempByteArray;
	struct TreeAsIsContext *pTreeAsIsContext;

	Log(">> WriteBuffer()\n");
	/*
	if (IsDebuggerPresent())
	{
		DebugBreak();
	}
	*/
	pTreeAsIsContext = (struct TreeAsIsContext *)context;
	rootElement = cpiRootElement(&iRc, parser);
	if (rootElement != CCI_NULL_ADDR)
	{
		tempByteArray.size = 100;
		tempByteArray.pointer = malloc(tempByteArray.size);
		ProcessNode(pTreeAsIsContext, parser, rootElement, &tempByteArray);
		free(tempByteArray.pointer);
	}
	Log("<< WriteBuffer()\n");
	return(0);
}

///////////////////////////////////////////////////////////
// ParserType
CciBool ParserType(CciParser *parser, CciContext* context)
{
	return(1);
}

///////////////////////////////////////////////////////////
// bipGetParserFactory
void *bipGetParserFactory()
{
	CciFactory*     factoryObject;
	int             rc;
	static CPI_VFT  vftable;			// Must be declared static
	
	Log("*** TreeAsIs Build: %s ***\n", __TIMESTAMP__);
	Log("%s", "Entered bipGetParserFactory()\n");

#ifdef _WIN32
	// Register the message catalog in the Windows NT/2000 registry.
	mqsixRegisterMessages(g_hModule, "MQSIV2_TREEASIS");
#endif

	/* Setup virtual function table with addresses of implementation functions */
	memset(&vftable, 0, sizeof(vftable));
	vftable.iFpCreateContext			= CreateContext;
	vftable.iFpDeleteContext			= DeleteContext;
	vftable.iFpParseBuffer				= ParseBuffer;
	vftable.iFpParseFirstChild			= ParseFirstChild;
	vftable.iFpParseLastChild			= ParseLastChild;
	vftable.iFpParseLeftSibling			= ParsePreviousSibling;
	vftable.iFpParseRightSibling		= ParseNextSibling;
	vftable.iFpWriteBuffer				= WriteBuffer;
	vftable.iFpIsHeaderParser			= ParserType;
	

	/* Create the message factory for this plug-in */
	factoryObject = cpiCreateParserFactory(&rc, L"TreeAsIsFactory");
	if (rc != CCI_SUCCESS)
	{
		Log("Failed to create parser factory: rc=%d\n", rc);
	}
	if (factoryObject)
	{

		/* Define the classes of message supported by the factory */
		cpiDefineParserClass(&rc, factoryObject, L"TREEASIS", &vftable);
		if (rc != CCI_SUCCESS)
		{
			Log("Failed to create a parser: rc=%d, %s\n", rc, ErrorToString(rc));
		}
		else
		{
			Log("Created the parser called %s\n", "TREEASIS");
		}
	}
	else
	{
		Log("Failed to create factory object in bipGetParserFactory()\n");
		/* Error recovery */ 
	}

	/* Return address of this factory object to the broker */
	return(factoryObject);
}
